--[[****************************************************************************
Mecaire's Macrotexter
******************************************************************************]]

--[[****************************************************************************
Declarations and Initialization
******************************************************************************]]

--------------------------------------------------------------------------------
-- "Constants"
--------------------------------------------------------------------------------
local DEFAULT_ALT = false;			-- Default selfcast settings: alt-selfcast
local DEFAULT_CTRL = false;         --                     ctrl-click selfcast
local DEFAULT_RGT = true;           --                    right-click selfcast

local HIDEGRID_DELAY = 0.2;         -- # secs to wait after ACTIONBAR_HIDEGRID to check for an update.
local PEW_INIT_DELAY = 5;           -- # secs to wait after PLAYER_ENTERING_WORLD to scan hotbars and initialize.
local SLOW_LOAD_DELAY = 1;			-- # secs to wait if player isn't loaded, before trying again
local IN_COMBAT_DELAY = 5;			-- # secs to wait if player is in combat during initialization, before trying again
local COMBAT_MSG_DELAY = 45;        -- # secs to wait between reminding the player the addon can't initialize when in combat

local MAX_ACTION_SLOTS = 120;       -- Maximum number of action slots

local ACTION_TYPE_SPELL = "spell";  -- Indicates a spell action slot
local ACTION_TYPE_ITEM = "item";    -- Indicates an item action slot

local NUM_BARS = 16;                -- Number of hotbars supported.

local MACRO_SELF = "SELFCAST";		-- Token included in macro template for selfcast spells
local MACRO_SPELLNR = "SPELLN";		-- Token included in macro template for spell name and rank

--------------------------------------------------------------------------------
-- Global variable declarations
--------------------------------------------------------------------------------
g_MMT_UpdateTimer = nil;            -- # secs to wait before checking the hotbars for updates.
g_MMT_InitTimer = nil;              -- # secs to wait before initial hotbar scans
g_MMT_CombatTimer = nil;			-- # secs to wait between displaying "in combat" messages when trying to initialize.
g_MMT_init = false;           		-- Set to true after the hotbars have been initialized the first time.
--------------------------------------------------------------------------------
-- Variable declarations
--------------------------------------------------------------------------------
local bUpdateNoCombat;				-- Set to true when an out-of-combat update is needed.

local oCurrentSlotData;        		-- The slot data corresponding with the current button settings.
local oSpellList;		            -- The spell list (see SpellList.lua).
local oMacroTemplate;				-- The macro templates by spell category (see SpellCategory.lua).

local oBars; 						-- Actionbar data for supported bars (See ActionBars.lua)
local oHeaders;						-- Actionbar headers (indexed by bar ID)

local mstrClass;            		-- The player's English class (all CAPS).

--[[****************************************************************************
MMT_OnLoad()

In the OnLoad() function, we do three things:
	1. Initialize saved variables
	2. Enable slash commands
	3. Register events
******************************************************************************]]
function MMT_OnLoad()

	-- Initialize saved variables.
	MMT_ALT_SELF = DEFAULT_ALT;
	MMT_CTRL_SELF = DEFAULT_CTRL;
	MMT_RGT_SELF = DEFAULT_RGT;
	
	-- Enable slash commands:  /macrotexter, /mmt
	SLASH_MACROTEXTER1 = "/macrotexter";
	SLASH_MACROTEXTER2 = "/mmt";
	SlashCmdList["MACROTEXTER"] = MMT_SlashCmd;

	-- Register events.
	this:RegisterEvent("ADDON_LOADED");			-- Loading saved variables
	this:RegisterEvent("PLAYER_ENTERING_WORLD");-- Initial hotbar scan
	this:RegisterEvent("ACTIONBAR_HIDEGRID");	-- Determining button changes
	this:RegisterEvent("LEARNED_SPELL_IN_TAB"); -- Automatic bar updates from trainer
	this:RegisterEvent("PLAYER_REGEN_ENABLED"); -- Leaving combat
		
end

--[[****************************************************************************
MMT_OnEvent()

******************************************************************************]]
function MMT_OnEvent()

	-----------------------------------------------------------------
	-- ACTIONBAR_HIDEGRID
	-- The player may have dragged a button on or off the bar.
	-- Set the update timer.
	-----------------------------------------------------------------
	if (event == "ACTIONBAR_HIDEGRID") then
        g_MMT_UpdateTimer = HIDEGRID_DELAY;
        
	-----------------------------------------------------------------
	-- ADDON_LOADED
	-- Load saved variables (selfcast settings).
	-----------------------------------------------------------------
	elseif ((event == "ADDON_LOADED") and (arg1 == "Macrotexter")) then

		-- Load saved selfcast settings.
		if MMT_ALT_SELF == DEFAULT_ALT
			and MMT_CTRL_SELF == DEFAULT_CTRL
			and MMT_RGT_SELF == DEFAULT_RGT then
			-- No changes - must be new user
		else 
			--MMT_ALT_SELF;	-- User selfcast settings:  alt selfcast
			--MMT_CTRL_SELF;  --                   ctrl-click selfcast
			--MMT_RGT_SELF;    --                  right-click selfcast
		end

		ShowMessage("Mecaire's Macrotexter loaded.  Initializing in ~"
			..PEW_INIT_DELAY.." seconds.");
		
		-- Display selfcast settings.
		SelfcastMessage("ALT");
		SelfcastMessage("CTRL");
		SelfcastMessage("RCLICK");
		
		-- Show slash commands.
		ShowMessage("   Slash commands: /mmt or /macrotexter");

	-----------------------------------------------------------------
	-- PLAYER_ENTERING_WORLD
	-- This event fires after login, /reloadui, and zoning.
	-- We wait for the player to be fully loaded, and then we
	-- scan the actionbars.
	-----------------------------------------------------------------
	elseif (event == "PLAYER_ENTERING_WORLD") then
	    if not g_MMT_init then g_MMT_InitTimer = PEW_INIT_DELAY; end;
	
	-----------------------------------------------------------------
	-- LEARNED_SPELL_IN_TAB
	-- The player learned a new spell.  The hotbar may have been 
	-- updated automatically by the trainer.  So we need to check it
	-- for updates.
	-----------------------------------------------------------------
	elseif (event == "LEARNED_SPELL_IN_TAB") then
		if g_MMT_init then MMT_UpdateButtons(false); end;
	
	-----------------------------------------------------------------
	-- PLAYER_REGEN_ENABLED
	-- The player is leaving combat.  Update the actionbars if anything
	-- changed while in combat.
	-----------------------------------------------------------------
	elseif (event == "PLAYER_REGEN_ENABLED") then	
		if g_MMT_init and bUpdateNoCombat then
			ShowMessage("Mecaire's Macrotexter is updating buttons now.");
			bUpdateNoCombat = false;
			MMT_UpdateButtons(false);
		end
	end
	
end

--[[****************************************************************************
MMT_DebugPrint(bShowMacros)

This function prints relevant internal data.  This includes the list of modified
spell buttons, current selfcast settings, tooltip state, and the states of
multi-page and multi-stance action bars.

If bShowMacros is true, the button macrotext will be printed.  This argument
is included because the output can be very lengthy.
******************************************************************************]]
function MMT_DebugPrint(bShowMacros)

	local strState;			-- The state of a state header.
	local strText;			-- Text to display.
	
	-- Print tooltip number of lines.
	ShowMessage("The hidden tooltip has "..MMT_Tooltip:NumLines().." lines.");

	-- Print the player's class.
	if mstrClass ~= nil then
		ShowMessage("Player Class: "..mstrClass);
	else
		ShowMessage("Player Class: nil");
	end

   	-- Print the states of multi-page and multi-stance action bars.
	if g_MMT_init then
		for iCount = 1, NUM_BARS do
			if (oBars[iCount].type == "multi_page")
				or (oBars[iCount].type == "stance_driven") then
				
				strState = oHeaders[oBars[iCount].id]:GetAttribute("state");
				if strState == nil then strState = "nil"; end;
				
				ShowMessage(oBars[iCount].id.." state: "..strState);
			end
		end
	else
		ShowMessage("(Not initialized - cannot display header states.)");
	end
	
	-- Print the list of modified spell buttons.
	if g_MMT_init then
		for iCount = 1, MAX_ACTION_SLOTS do
			if oCurrentSlotData[iCount].spell_name ~= nil then
				ShowMessage("Slot "..iCount..": "
					..MMT_SpellNR(oCurrentSlotData[iCount].spell_name
					, oCurrentSlotData.spell_rank)
					..", type "..oCurrentSlotData[iCount].spell_type);
			end
		end
	else
		ShowMessage("(Not initialized - cannot display modified buttons.)");
	end
	
	-- Print available categories.
	if g_MMT_init then
	    for intIndex, oCategory in pairs(oMacroTemplate) do
			ShowMessage("Category ["..intIndex.."] : "..oCategory.template);
		end
	else
	    ShowMessage("(Not initialized - cannot display categories.)");
	end
	
	-- Print button macrotext if specified.
	if bShowMacros and g_MMT_init then
		for iCount = 1, NUM_BARS do
			if oBars[iCount].active then
				for jCount = oBars[iCount].min_slot, oBars[iCount].max_slot do
					strText = MMT_GetMacrotext(jCount, oBars[iCount]);
					if strText ~= nil then
						ShowMessage(oBars[iCount].id.." slot "..jCount..": "..strText);
					end
				end
			end
		end
	end
	
	-- Print current selfcast settings.
	SelfcastMessage("ALT");
	SelfcastMessage("CTRL");
	SelfcastMessage("RCLICK");
			
end

--[[****************************************************************************
SelfcastMessage(strModifier)

Shows a message indicating whether the specified selfcasting is on or off.

******************************************************************************]]
function SelfcastMessage(strModifier)
	
	local strMessage;		-- The message to display.
	local bValue;			-- The value to display.
	
	if strModifier == "ALT" then
		strMessage = "Alt-";
		bValue = MMT_ALT_SELF;
	elseif strModifier == "CTRL" then
		strMessage = "Ctrl-click ";
		bValue = MMT_CTRL_SELF;
	elseif strModifier == "RCLICK" then
		strMessage = "Right-click ";
		bValue = MMT_RGT_SELF;
	end
		
	ShowMessage("   "..strMessage.."selfcasting is "..TrueFalseToOnOff(bValue));
end

--[[****************************************************************************
MMT_SlashCmd()

This function is the slash command handler.

******************************************************************************]]
function MMT_SlashCmd(msg)

	local strCommand;	
	
	-- Convert the command line arguments to a table.
	local slash_table = MsgToTable(msg);

	strCommand = slash_table[1];
	if strCommand ~= nil then
		strCommand = strupper(strCommand);        -- First command
	end

    ----------------------------------------------------------------------------
    -- /macrotexter
	-- No command line argument:  Display a message.
	----------------------------------------------------------------------------
	if strCommand == nil then
		ShowMessage("Please use the Addons option at the character login"
			.." screen to toggle"
			.." Macrotexter.  Type '/macrotexter help' to view commands.");
	
	----------------------------------------------------------------------------
	-- /macrotexter debug [macros]
	-- "Debug" command: Display the internal data to the screen.
	-- If "macros" is included, displays the button macrotext.
	----------------------------------------------------------------------------
	elseif strCommand == "DEBUG" then
		if slash_table[2] ~= nil then
			if strupper(slash_table[2]) == "MACROS" then
				MMT_DebugPrint(true);
			else
				MMT_DebugPrint(false);
			end
		else
		    MMT_DebugPrint(false);
		end
	
	----------------------------------------------------------------------------
	-- /macrotexter alt|ctrl|rclick
	-- Toggles selfcasting and updates buttons accordingly.
	----------------------------------------------------------------------------	
	elseif strCommand == "ALT"
		or strCommand == "CTRL"
		or strCommand == "RCLICK" then
		
		ToggleSelfcasting(strCommand);
		if g_MMT_init then
			MMT_UpdateButtons(false);
		end
	
	----------------------------------------------------------------------------
	-- /macrotexter cat #
	-- Toggle macrotexter for the given category.
	-- TODO: Finish this function
	----------------------------------------------------------------------------
--[[	elseif strCommand == "CAT" then
        if MMT_IsValidCategory(slash_table[2]) then
			MMT_ToggleCategory(slash_table[2]);
			if g_MMT_init then
			    MMT_UpdateButtons(false);
			end
        else
            ShowMessage("Invalid category.");
        end
]]

	----------------------------------------------------------------------------
	-- /macrotexter update
	-- Sloppy manual update (intended for Bartender3 settings changed).
	-- Currently remarked - use /reloadui instead :(
	--      (Since MMT_UpdateButtons() won't always recognize
	--      bar structure changes properly)
	-- TODO: Finish hooking job & remove this / cleanup
	----------------------------------------------------------------------------
--[[	elseif strCommand == "UPDATE" then
		if g_MMT_int then
			if Bartender3 then MMT_Bartender3Init(); end;
			MMT_UpdateButtons(false);
		else
		    ShowMessage("(Macrotexter isn't initialized - can't update.)");
		end
]]

	----------------------------------------------------------------------------
	-- /macrotexter off
	-- Disable macrotexter.
	----------------------------------------------------------------------------
--[[	elseif strCommand == "OFF" then
		-- TODO: Finish
]]
	----------------------------------------------------------------------------
	-- /macrotexter on
	-- Enable macrotexter.
	----------------------------------------------------------------------------
--[[	elseif strCommand == "ON" then
		-- TODO: Finish
]]

	----------------------------------------------------------------------------
	-- /macrotexter help
	-- Display the available slash commands.
	----------------------------------------------------------------------------	
	elseif strCommand == "HELP" then
		ShowMessage("/macrotexter debug [macros] : Displays internal data.");
		ShowMessage("/macrotexter alt    : Toggles alt-selfcasting.");
		ShowMessage("/macrotexter ctrl   : Toggles ctrl-click-selfcasting.");
		ShowMessage("/macrotexter rclick : Toggles right-click-selfcasting.");
		--ShowMessage("/macrotexter cat #  : Toggle a category on or off.");
		--ShowMessage("/macrotexter off|on : Turn Macrotexter on or off.");
		ShowMessage("/macrotexter help   : Displays this message.");

	----------------------------------------------------------------------------
	-- Invalid command
	----------------------------------------------------------------------------	
	else
		ShowMessage("Invalid command.  Type '/macrotexter help' for a"
			.." list of valid commands.");
	end	
end

--[[****************************************************************************
MMT_OnUpdate(elapsed)

This function handles the two timers which may be counting down from the
OnUpdate event.

When g_MMT_InitTimer has finished, we need to scan hotbars and make our initial
button changes.

When g_MMT_UpdateTimer has finished, we need to check for updates to the
hotbars.
******************************************************************************]]
function MMT_OnUpdate(elapsed)

	----------------------------------------------------------------------------
	-- Countdown to Initialization
	-- 	We must not have already initialized.
	-- 	We must ensure the player is fully loaded before initializing.
	--  We continue to wait for 1 second intervals until the player is loaded,
	--  if necessary.
	--  We must wait until the player is out of combat (such as if a druid
	--  with Furor logged in in shapeshift form).
	--  We trigger the initialization procedure.
	----------------------------------------------------------------------------
	if (not g_MMT_init) and (g_MMT_InitTimer ~= nil) then
	    g_MMT_InitTimer = g_MMT_InitTimer - elapsed;

		if g_MMT_CombatTimer ~= nil then                    -- For in-combat message display
			g_MMT_CombatTimer = g_MMT_CombatTimer - elapsed;
		end

		-- Ensure the player is fully loaded before initializing.
		if g_MMT_InitTimer < 0 then
			g_MMT_InitTimer = nil;
			
			if not PlayerLoaded() then
				ShowMessage("Mecaire's Macrotexter is waiting a moment"
					.." while you finish loading.");
			    g_MMT_InitTimer = SLOW_LOAD_DELAY;    -- Wait one more second.
			
			-- We cannot initialize if the player is in combat, because the
			-- UI doesn't allow us to set up and attach state headers and
			-- adjust button attributes.
			elseif UnitAffectingCombat("player") then
				
				-------------Message Display Only-------------------------------
				-- Only display the in-combat delay message once per minute.
				if g_MMT_CombatTimer == nil then
					g_MMT_CombatTimer = -1; --Initial message - show immediately
				end
				
				if g_MMT_CombatTimer < 0 then
					g_MMT_CombatTimer = COMBAT_MSG_DELAY;
					ShowMessage("(You are in combat.  Mecaire's Macrotexter must"
						.." wait to initialize...)");
				end
				----------------------------------------------------------------
				
				g_MMT_InitTimer = IN_COMBAT_DELAY;	  -- Wait five more seconds.
			else
				MMT_InitButtons();
			end
	    end
	
    ----------------------------------------------------------------------------
	-- Countdown to Update
	-- We must have already initialized.
	-- We trigger the button update procedure.
	----------------------------------------------------------------------------
	elseif g_MMT_init and g_MMT_UpdateTimer ~= nil then
	    g_MMT_UpdateTimer = g_MMT_UpdateTimer - elapsed;
	    if g_MMT_UpdateTimer < 0 then
	        g_MMT_UpdateTimer = nil;
	        
			MMT_UpdateButtons(false);
	        
		end
	end
end

--[[****************************************************************************
ToggleSelfcasting(strModifier)

This function toggles the selfcasting setting for the given modifier.

strModifier must be one of: "ALT", "CTRL", or "RCLICK".

******************************************************************************]]
function ToggleSelfcasting(strModifier)

	-- Toggle selfcasting for the given modifier.
	if strModifier == "ALT" then
		MMT_ALT_SELF = not MMT_ALT_SELF;
	elseif strModifier == "CTRL" then
		MMT_CTRL_SELF = not MMT_CTRL_SELF;
	elseif strModifier == "RCLICK" then
		MMT_RGT_SELF = not MMT_RGT_SELF;
	end
	
	-- Display selfcasting message.
	SelfcastMessage(strModifier);
end

--[[****************************************************************************
MMT_IsValidCategory(value)

Returns True if the given value is a valid spell category.  Returns false
otherwise.
******************************************************************************]]
function MMT_IsValidCategory(value)
	return (oMacroTemplate[value] ~= nil);
end

--[[****************************************************************************
MMT_IsSelfCastCat(value)

Returns true if the given value is for a macro template category including
the Macrotexter selfcast token.  Returns false otherwise.
******************************************************************************]]
function MMT_IsSelfcastCat(value)
	if MMT_IsValidCategory(value) then
		return (string.find(oMacroTemplate[value].template, MACRO_SELF) ~= nil);
	else
	    return false;
	end
end

--[[****************************************************************************
ScanSlotsToTable()

This function scans the player's action slots and enters data into a table.
Slot data is only recorded if the slot's button will be altered in some way
by this addon.  The resulting table is returned.
******************************************************************************]]
function ScanSlotsToTable()

	local oTable = {};              -- The table to be returned.
	local strType = "";             -- Action slot type
	local strSpellName = "";        -- Spell name from tooltip
	local strSpellRank = "";        -- Spell rank from tooltip
	local intSpellType;             -- Spell type (from SpellList.lua)
	
	-- Initialize selfcast values.
	oTable.selfcast = {};
	oTable.selfcast["ALT"] = MMT_ALT_SELF;
	oTable.selfcast["CTRL"] = MMT_CTRL_SELF;
	oTable.selfcast["RCLICK"] = MMT_RGT_SELF;
	
	-- Prepare our tooltip.  It occasionally becomes unresponsive (such as
	-- during a slow load) if we do not reset the owner to WorldFrame here.
	MMT_Tooltip:SetOwner(WorldFrame, "ANCHOR_NONE");
	
	-- For each action slot:
	for iCount = 1, MAX_ACTION_SLOTS do

        -- Initialize the table entry.
		oTable[iCount] = {};

		-- We're only interested in spell and item action slots:
		strType = GetActionInfo(iCount);
		if strType == ACTION_TYPE_SPELL or strType == ACTION_TYPE_ITEM then
		
			-- Read the spell name and rank using our tooltip.
			MMT_Tooltip:ClearLines();
			MMT_Tooltip:SetAction(iCount);
			strSpellName = MMT_TooltipTextLeft1:GetText();
			strSpellRank = MMT_TooltipTextRight1:GetText();
			
			-- If this spell is in our spell table (see SpellList.lua),
			-- and its category is defined (see SpellCategory.lua),
			-- then add it to our slot table.
			intSpellType = oSpellList[strSpellName];
			if MMT_IsValidCategory(intSpellType) then
			
				-- Only record information if this category is active:
				if oMacroTemplate[intSpellType].active then
					oTable[iCount].spell_name = strSpellName;   -- Spell name
					oTable[iCount].spell_rank = strSpellRank;   -- Spell rank
					oTable[iCount].spell_type = intSpellType;   -- Spell type
				end
			end
		end
	end
	return oTable;
end

--[[****************************************************************************
MMT_InitButtons()

This function does the initial scan of hotbars and buttons.  This is where the
state headers will be attached to the buttons.

******************************************************************************]]
function MMT_InitButtons()

	local oBar;                     -- A reference to one actionbar's data.
	local oButton;					-- A reference to the button for a slot.
	
	-- Get the player's class and initialize the spell table and categories.
	_, mstrClass = UnitClass("player");
    oSpellList = MMT_GetSpellList(mstrClass);
	oMacroTemplate = MMT_MakeCategories();
	
	-- Initialize the actionbar data.
	oBars = MMT_GetActionbars();

	-- Create state headers.
	MMT_CreateStateHeaders();

	-- Get Bartender3 actionbar information, if Bartender3 is installed.
	if Bartender3 then MMT_Bartender3Init(); end;

	-- Scan the player's action slots to oCurrentSlotData.
	oCurrentSlotData = ScanSlotsToTable();

	-- Loop through each supported actionbar and its slots.
	for iCount = 1, NUM_BARS do

		oBar = oBars[iCount];

		-- Only set up bars that are active.
		if oBar.active then
			-- Go through the slot data and attach state headers as needed.
			for jCount = oBar.min_slot, oBar.max_slot do
				if oCurrentSlotData[jCount].spell_name ~= nil then
				
            		-- Get a reference to the button and attach it to the
					-- state header.
					oButton = MMT_GetButton(jCount, oBar);
					AttachHeaderToButton(oBar.type, oHeaders[oBar.id], oButton);
	    		end
			end
		end
	end

	-- Now that the headers are set up, update the type, macrotext, etc. for 
	-- the actionbuttons.
	MMT_UpdateButtons(true);
	
	-- Set the initialized status to true:
	g_MMT_init = true;
	
	if bUpdateNoCombat then
		-- We shouldn't get here because this function is protected from calls
		-- in combat, but it's possible the player may have entered combat
		-- during UpdateButtons().
		ShowMessage("Mecaire's Macrotexter will initialize when out of combat.");
	else
		ShowMessage("Mecaire's Macrotexter successfully initialized.");
	end
	
end

--[[****************************************************************************
MMT_Bartender3Init()

Gets Bartender3 statemaps and sets slot values, number of buttons per bar.
******************************************************************************]]
function MMT_Bartender3Init()

	local oBT3Bar;
	local oState;       -- Default state object.
	 
	
	for iCount = 1, NUM_BARS do
		oBar = oBars[iCount];
		if oBar.type == "bartender" then
			
			oBT3Bar = Bartender3:GetBarObject(oBar.index, false);
			oBar.active = (oBT3Bar ~= nil);

			-- TODO: Hook
			--oBT3Bar:RefreshLayout = MMT_BT3_RefreshLayout; -- oBar.index, oBT3Bar
			
			if oBar.active then
				
				-- Try to hook (TODO) RefreshLayout for this bar.
				--hooksecurefunc({oBT3Bar}, "RefreshLayout", MMT_BT3_RefreshLayout);
				
				-- Get the statemap from the bar frame
				oBar.statemap = MMT_StatemapToTable(
					oBT3Bar.frame:GetAttribute("statebutton"));
			
				-- Every bar can contain any of the possible action slots.
				oBar.min_slot = 1;
				oBar.max_slot = MAX_ACTION_SLOTS;
			
				-- Get the number of buttons for this bar.
				oBar.num_buttons = oBT3Bar.numbuttons;
			end
		end
	end
	
end

--[[****************************************************************************
MMT_UpdateButtons(bInit)

This function updates the buttons when they are moved on or off the bar, or
automatically updated by training.

bInit indicates whether to skip taking a new snapshot of the actionbars.
bInit should always be false unless this function is called from 
MMT_InitButtons() for the first time.

******************************************************************************]]
function MMT_UpdateButtons(bInit)

	local tNew;				-- The "new" table of actionbar slots.
	local oBar;				-- A reference to the data about the actionbar.
	local bSelfcastChanged; -- True if selfcast settings changed since last time.
	local bChanged = false; -- True if anything changed requiring an update.
	
	-- We cannot perform an update if the player is in combat.
	-- Set bUpdateNoCombat if the player is in combat, and use
	-- it and bChanged to determine if we need to display a message.
	if UnitAffectingCombat("player") then
		bUpdateNoCombat = true;
	end
	
	-- If not initializing, take a snapshot of actionbar slots for comparison.
	if not bInit then
		tNew = ScanSlotsToTable();
	end

	-- Determine whether selfcasting changed.
	bSelfcastChanged = bInit;	  -- If we're initializing, consider it changed.
	if not bSelfcastChanged then
		bSelfcastChanged = (tNew.selfcast["ALT"] ~= oCurrentSlotData.selfcast["ALT"]
			or tNew.selfcast["CTRL"] ~= oCurrentSlotData.selfcast["CTRL"]
			or tNew.selfcast["RCLICK"] ~= oCurrentSlotData.selfcast["RCLICK"]);
	end

	-- For each actionbar:
	for iCount = 1, NUM_BARS do
		
		oBar = oBars[iCount];
		
		-- Don't update unused bars.
		if oBar.active then
			-- For each action slot used by this actionbar:
			for jCount = oBar.min_slot, oBar.max_slot do
				
				-- If we're initializing, always update (unless in combat).
				if bInit then
					if oCurrentSlotData[jCount].spell_name ~= nil then
						bChanged = true;
						if not bUpdateNoCombat then
							MMT_SetMacrotext(jCount, oBar, oCurrentSlotData[jCount], true);
						end
					end
				
				-- If we're not initializing, check for changes.
				else
					-- If there is nothing in the "new" slot but something in the "old"
					-- slot, remove the changes we'd made to the button.
					if tNew[jCount].spell_name == nil
						and oCurrentSlotData[jCount].spell_name ~= nil then
					
					    bChanged = true;
					
					    if not bUpdateNoCombat then
							MMT_SetMacrotext(jCount, oBar, tNew[jCount], false);
					    end
					    
					-- If what is in the "new" slot doesn't match the "old" slot, or
					-- if selfcasting changed and this is a beneficial spell => update
					-- it.
					elseif tNew[jCount].spell_name ~= oCurrentSlotData[jCount].spell_name
						or tNew[jCount].spell_rank ~= oCurrentSlotData[jCount].spell_rank
						or (bSelfcastChanged 
						 and MMT_IsSelfcastCat(tNew[jCount].spell_type)) then
				
				        bChanged = true;
				        
				        if not bUpdateNoCombat then
							MMT_SetMacrotext(jCount, oBar, tNew[jCount], true);
						end
						
					end
				end
			end
		end
	end
	
	-- Display a message if the player was in combat and there are changes
	-- to be made.
	bUpdateNoCombat = bUpdateNoCombat and (bChanged or bSelfcastChanged);
	if bUpdateNoCombat then
		ShowMessage("Mecaire's Macrotexter cannot make changes in combat."
		.." Until you leave combat, your action buttons may not function properly.");

	-- If we're out of combat, not initializing, and any changes were made,
	-- apply them to the current table data.
	elseif (not bInit) and (bChanged or bSelfcastChanged) then
	    oCurrentSlotData = tNew;
	end
	
end

--[[****************************************************************************
MMT_CreateStateHeaders()

Creates state headers.

The actionbar data table oBars must already be filled.
******************************************************************************]]
function MMT_CreateStateHeaders()
    
    local oBar;			-- A reference to one actionbar's data.
    
    oHeaders = {};		-- Initialize headers table.
    
	for iCount = 1, NUM_BARS do
		
		oBar = oBars[iCount];
	
		if oBar.type == "multi_page" or oBar.type == "stance_driven" then
	
			oHeaders[oBar.id] = CreateFrame("Frame", oBar.id.."DriverFrame"
				, UIParent, "SecureStateDriverTemplate");
			oHeaders[oBar.id]:SetPoint("CENTER", 0, 0);
			oHeaders[oBar.id]:SetWidth(2);
			oHeaders[oBar.id]:SetHeight(2);
			oHeaders[oBar.id]:SetFrameStrata("HIGH");

			if oBar.type == "multi_page" then
				oHeaders[oBar.id]:SetAttribute("statemap-actionbar-1", "1");
				oHeaders[oBar.id]:SetAttribute("statemap-actionbar-2", "2");
				oHeaders[oBar.id]:SetAttribute("statemap-actionbar-3", "3");
				oHeaders[oBar.id]:SetAttribute("statemap-actionbar-4", "4");
				oHeaders[oBar.id]:SetAttribute("statemap-actionbar-5", "5");
				oHeaders[oBar.id]:SetAttribute("statemap-actionbar-6", "6");
			else        -- stance_driven
				oHeaders[oBar.id]:SetAttribute("statemap-stance-0", "0");
				oHeaders[oBar.id]:SetAttribute("statemap-stance-1", "1");
				oHeaders[oBar.id]:SetAttribute("statemap-stance-2", "2");
				oHeaders[oBar.id]:SetAttribute("statemap-stance-3", "3");
				oHeaders[oBar.id]:SetAttribute("statemap-stance-4", "4");
				oHeaders[oBar.id]:SetAttribute("statemap-stance-5", "5");
			end
		end
	end
end

--[[****************************************************************************
AttachHeaderToButton(oType, oHeader, oButton)

This function attaches the state header to an actionbutton.

Only buttons on actionbars of type "multi_page" and "stance_driven use state
headers.

******************************************************************************]]
function AttachHeaderToButton(strType, oHeader, oButton)

	local intState = 0;			-- Initial state for the header.

	-- Don't set up the state header if the actionbar is not the right type.
	if strType ~= "multi_page" and strType ~= "stance_driven" then
		return;
	end

	--When an attribute of this button is changed, tell it to update
	-- itself (this includes state changes from the driver)
	oButton:SetScript("OnAttributeChanged", ActionButton_Update);
	
	if strType == "multi_page" then
		oButton:SetAttribute("statebutton", "1:S1;2:S2;3:S3;4:S4;5:S5;6:S6");
	elseif strType == "stance_driven" then
	    oButton:SetAttribute("statebutton", "0:S0;1:S1;2:S2;3:S3;4:S4;5:S5");
	end

	-- Add the button to the header as a child.
	-- The following line causes druid/warrior BonusBar buttons affected to disappear from the screen.
	-- It also seems to be unnecessary.  So it's remarked out.
--	oHeader:SetAttribute("addchild", oButton)
    
	-- Associate the state header with the button, per Kushiel!
	oButton:SetAttribute("stateheader", oHeader)

    -- Set the initial state.
	if strType == "multi_page" then
		intState = GetActionBarPage();
	elseif strType == "stance_driven" then
		intState = GetShapeshiftForm(true);
	end
	oHeader:SetAttribute("state", intState);
	
end

--[[****************************************************************************
MMT_GetButton(intSlot, oBar)

Returns the button object for the given action slot number and actionbar.

The actionbutton index is calculated by simple modulus arithmetic.
Actionbuttons are numbered 1 to oBar.num_buttons.
Action slots can be numbered 1 to 120.
So, if oBar.num_buttons is 12, then action slots 1, 13, 25, etc. map to
 actionbutton index 1.
******************************************************************************]]
function MMT_GetButton(intSlot, oBar)

	local intIndex = 0;		-- The index for this button.

	-- Calculate the index.
	intIndex = intSlot % oBar.num_buttons;
	if intIndex == 0 then intIndex = oBar.num_buttons; end;

	-- For Bartender bars, the button index must be adjusted by the bar number (index).
	if oBar.type == "bartender" then
	    intIndex = intIndex + (oBar.num_buttons * (oBar.index - 1));
	end

	-- Return the actionbutton.
	return getglobal(oBar.button_name..intIndex);

end


--[[****************************************************************************
MMT_SpellNR(strName, strRank)

Puts the spell name and rank together and returns the spell name and rank
as it appears on UI tooltips.  (NR = Name and Rank)
******************************************************************************]]
function MMT_SpellNR(strName, strRank)

	if strRank == nil then
		return strName;
	else
		return strName.."("..strRank..")";
	end
end

--[[****************************************************************************
MMT_MakeMacro(oSpell)

Returns a macro created for the given spell object
	(oSpell.spell_name, oSpell.spell_rank, oSpell.spell_type)

******************************************************************************]]
function MMT_MakeMacro(oSpell)

	local strTemplate;		-- The macro template for this type of spell.
	local strSelf;			-- Selfcast macro text.
	local strSpellNR;		-- Spell name and rank.
	local strMacro;			-- The macro result.
	
	-- Determine the macro template for the spell.
	strTemplate = oMacroTemplate[oSpell.spell_type].template;
	
	-- Put the spell name and rank (if any) together.
	strSpellNR = MMT_SpellNR(oSpell.spell_name, oSpell.spell_rank);
	
	-- "SELF" in the macro template needs replaced by appropriate selfcast commands.
	strSelf = "";
	if MMT_ALT_SELF then
		strSelf = strSelf.."[modifier:alt,target=player] "..MACRO_SPELLNR.."R;";
	end
	if MMT_CTRL_SELF then
		strSelf = strSelf.."[modifier:ctrl,button:1,target=player] "..MACRO_SPELLNR.."R;";
	end
	if MMT_RGT_SELF then
		strSelf = strSelf.."[button:2,target=player] "..MACRO_SPELLNR.."R;";
	end
	
	-- Replace "SELF" token first.
	strMacro = gsub(strTemplate, MACRO_SELF, strSelf);
	
	-- Finally, replace the "SPELL-NR" token------------------------------------

	-- TODO: Finish
	-- SPELLN# - A particular spell rank
	for strTemp in string.gmatch(strMacro, MACRO_SPELLNR.."%d") do
		strRank = MMT_SpellNR(oSpell.spell_name
			, "Rank "..string.sub(strTemp, string.find(strTemp, "%d")));
		strMacro = gsub(strMacro, MACRO_SPELLNR.."%d", strRank);
	end
	
	-- SPELLNR
	strMacro = gsub(strMacro, MACRO_SPELLNR.."R", strSpellNR);
	
	return strMacro;
end

--[[****************************************************************************
MMT_SetMacrotext(intSlot, oBar, bEnable)

Sets the button type and macrotext for the given slot and bar, if bEnable is
true.

If bEnable is false, undoes any changes.

******************************************************************************]]
function MMT_SetMacrotext(intSlot, oBar, oSpell, bEnable)

	local strTypeAttr;			-- Type attribute name.
	local strTypeValue;			-- Type attribute value.
	local strMacroAttr;			-- Macrotext attribute name.
	local strMacroValue;		-- Macrotext attribute value.
	local index;				-- Index
	local oState;				-- State object (Bartender only).
	
	local oButton;				-- The button for this action slot and bar.
	
	----------------------------------------------------------------------------
	-- Determine the attribute names.
	-- If the actionbar is multi-page or stance-driven, these include the
	-- state name.
	----------------------------------------------------------------------------
	strTypeAttr, strMacroAttr = MMT_TypeAndMacroAttr(intSlot, oBar);

	----------------------------------------------------------------------------
	-- Determine the attribute values.
	----------------------------------------------------------------------------
	if not bEnable then
		strTypeValue = "action";
		strMacroValue = "";
	else
		strTypeValue = "macro";
		strMacroValue = MMT_MakeMacro(oSpell);
	end

    ----------------------------------------------------------------------------
    -- Finally, set the button attributes.
	----------------------------------------------------------------------------
	oButton = MMT_GetButton(intSlot, oBar);					-- Get button.

	-- Bartender bars
	if oBar.type == "bartender" then

		-- If any state is mapped to this action slot, set the macrotext for
		-- that state.
		for index, oState in pairs(oBar.statemap) do
			if oButton:GetAttribute("*action-"..oState.name) == intSlot then
				oButton:SetAttribute("*"..strTypeAttr.."-"..oState.name, strTypeValue);
	    		oButton:SetAttribute("*"..strMacroAttr.."-"..oState.name, strMacroValue);
			end
		end

		-- Update macrotext if this is a default action (such as for stateless bars).
		if oButton:GetAttribute("action") == intSlot then
			oButton:SetAttribute("*"..strTypeAttr, strTypeValue);
			oButton:SetAttribute("*"..strMacroAttr, strMacroValue);
		end

	-- Non-bartender bars
	else
		oButton:SetAttribute(strTypeAttr, strTypeValue);        -- Set type.
		oButton:SetAttribute(strMacroAttr, strMacroValue);      -- Set macrotext.
	end
	
end

--[[****************************************************************************
MMT_GetMacrotext(intSlot, oBar)

Gets the button macrotext for the given slot and bar.

This is the actual, current macrotext in the button.  It is not calculated from
stored information.
******************************************************************************]]
function MMT_GetMacrotext(intSlot, oBar)
	
	local strTypeAttr;		-- Type attribute name.
	local strMacroAttr;		-- Macrotext attribute name.
	local oButton;			-- An actionbutton object.
	local index;			-- Index.
	local oState;			-- State object (bartender bars)
	local temp = "";		-- temp
	local strStates = "States";	-- State list
	local strMacro = "";	-- Macrotext
	local strStateMacro;    -- temp: macrotext for a particular state
	
	-- Get the type and macrotext attribute names.
	strTypeAttr, strMacroAttr = MMT_TypeAndMacroAttr(intSlot, oBar);
	
	-- Get the action button.
	oButton = MMT_GetButton(intSlot, oBar);

	-- Read the button's macrotext.
	
	-- Bartender bars:
	if oBar.type == "bartender" then
		for index, oState in pairs(oBar.statemap) do
			-- For bartender bars we return a list of states followed by the
			-- macro(s) associated with those states.
			if oButton:GetAttribute("*action-"..oState.name) == intSlot then
				strStateMacro = oButton:GetAttribute("*"..strMacroAttr.."-"..oState.name);
				if strStateMacro ~= nil then
					strStates, strMacro, temp = MakeBT3DebugText(strStates
						, " "..oState.name, strMacro, strStateMacro, temp);
				end
			end
		end

		-- Default action
		if oButton:GetAttribute("action") == intSlot then
			if oButton:GetAttribute(strMacroAttr) ~= nil then
				strStates, strMacro, temp = MakeBT3DebugText(strStates
					, " default", strMacro, oButton:GetAttribute(strMacroAttr)
					, temp);
			end
		end

		if strMacro == "" then
			return nil;
		else
		    return strStates..temp;
		end
	
	-- Non-Bartender bars
	else
		return oButton:GetAttribute(strMacroAttr);
	end
	
end

--[[****************************************************************************
MakeBT3DebugText()

This function helps create text to be returned by MMT_GetMacroText,
for Bartender3 buttons.

******************************************************************************]]
function MakeBT3DebugText(strStateList, strStateAppend, strLastMacro
	, strThisMacro, strMacroList)

	strStateList = strStateList..strStateAppend;
	if strLastMacro ~= strThisMacro then
		strLastMacro = strThisMacro;
		strMacroList = strMacroList.."\n"..strLastMacro;
	end
	
	return strStateList, strLastMacro, strMacroList;
	
end

--[[****************************************************************************
MMT_TypeAndMacroAttr(intSlot, oBar)

This is an auxilliary function intended to be called by MMT_GetMacrotext
and MMT_SetMacrotext.

It returns the type and macro attribute names for the given slot and bar.

If the actionbar is multi-page or stance-driven, these include the
state name.

******************************************************************************]]
function MMT_TypeAndMacroAttr(intSlot, oBar)
	
	local strType;			-- The "type" attribute name, to be returned.
	local strMacro;         -- The "macrotext" attribute name, to be returned.
	local intPage;			-- The action page number.
	local intStance;        -- The stance number.

	if oBar.type == "multi_page" then
		intPage = MMT_PageForSlot(intSlot, oBar);
		strType = "*type-S"..intPage;
		strMacro = "*macrotext-S"..intPage;
	elseif oBar.type == "stance_driven" then
	    intStance = MMT_StanceForSlot(intSlot, oBar, mstrClass);
	    strType = "*type-S"..intStance;
	    strMacro = "*macrotext-S"..intStance;

	-- Bartender:  Each slot may have many states.  Let the calling function
	-- append the states.
	elseif oBar.type == "bartender" then
		strType = "type";
		strMacro = "macrotext";

	else
	    strType = "*type*";
	    strMacro = "*macrotext*";
	end
	
	return strType, strMacro;
end

--[[****************************************************************************
MMT_StatemapToTable(strStatemap)

Converts a statemap string to a table of values and returns the table.
******************************************************************************]]
function MMT_StatemapToTable(strStatemap)

	local t = {};   	-- The table to be returned.
	local strVal;    	-- Temp string
	local strName;    	-- Temp string
	local oState;		-- Temp state object
	
	-- The Bartender3 statemap format is #1:State1;#2:State2;..#n:Staten;
	for strVal, strName in string.gmatch(strStatemap, "(%w+):(%w+);") do

    	-- Create a state object for this item.
		oState = {};
		oState.value = strVal;
		oState.name = strName;

		-- Store the state object in the table.
		tinsert(t, oState);
 	end
	
	return t;
end


-- TODO: Hook, Clean up
--[[function MMT_BT3_RefreshLayout()
	ShowMessage("Successful hook, RefreshLayout() - bar ");
	return oBT3Bar:RefreshLayout();
end
]]

--[[****************************************************************************
MMT_MakeCategories()

Gets the spell categories (see SpellCategory.lua) and adds an "active" field
to each one.  "Active" defaults to True.

******************************************************************************]]
function MMT_MakeCategories()

	local tOrig = MMT_GetSpellCategories();
	local tReturn = {};
	
	for intIndex, oCat in pairs(tOrig) do
	    tReturn[intIndex] = {};
	    tReturn[intIndex].template = oCat;
		tReturn[intIndex].active = true;
	end

	return tReturn;
end

--[[****************************************************************************
MMT_ToggleCategory(intIndex)

Toggles whether the given category is active.  Assumes the given category is
valid.

******************************************************************************]]
function MMT_ToggleCategory(intIndex)
	oMacroTemplate[intIndex].active = not oMacroTemplate[intIndex].active;
end